/* misc utility
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: misc.c,v 1.7 2006/10/18 09:11:23 gondow Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#include "misc.h"
#include "tef.h"

#define  ESC_FORM_FEED   "<![CDATA[&#x0C;]]>"
/* ==================================================================== */
/* ==================================================================== */
/* ==================================================================== */
char *
chop (char *s)
{
    if (s == NULL)
	return NULL;
    s [strlen (s) - 1] = '\0';
    return s;
}

char *
chop_nl (char *s)
{
    return chop_char (s, '\n');
}

char *
chop_char (char *s, int c)
{
    int len;
    if (s == NULL)
	return NULL;
    len = strlen (s);
    if (s [len - 1] == c)
	s [len - 1] = '\0';
    return s;
}

int
count_char (char *s, int c)
{
    int n = 0;
    char *p = s;

    while (1) {
	p = strchr (p, c);
	if (p == NULL) {
	    break;
	} else {
	    n++;
	    p++;
	}
    }
    return n;
}

char *
substring (char *s, int from, int to)
{
    char *p;

    assert (s != NULL);
    assert (0 <= from && from <= to);
    p = emalloc (to - from + 1);
    memcpy (p, s + from, to - from);
    p [to - from] = '\0';
    return p;
}

int
mkdirhier (char *dir, mode_t mode)
{
    char *slash = dir;
    struct stat sbuf;
    int ret;
    
    if (*slash == '/')
	slash++;

    while (1) {
	slash = strchr (slash, '/');
	if (slash != NULL)
	    *slash = '\0';
	
	ret = stat (dir, &sbuf);
	if (ret == -1) {
	    if (errno != ENOENT) { perror ("stat"); return -1; }
	    ret = mkdir (dir, mode);
	    if (ret == -1 && errno != EEXIST) { perror ("mkdir"); return -1; }
	} else if (!S_ISDIR(sbuf.st_mode)) {
	    fprintf (stderr, "%s: %s\n", dir, strerror (ENOTDIR));
	    return -1;
	}

	if (slash == NULL)
	    break;
	*slash++ = '/';
    }
    return 0;
}

FILE *
fopen_with_mkdir (const char *path, const char *mode)
{
    int ret;
    char *path2 = estrdup (path);
    FILE *fp = fopen (path, mode);

    if (fp == NULL && errno == ENOENT) {
	char *s = strrchr (path2, '/');
	if (s != NULL)
	    *s = '\0';
#if 1
	ret = mkdirhier (path2, 0777);
#else
	{
	    char *cmd;
	    asprintf (&cmd, "mkdir -p %s", path2);
	    ret = system (cmd);
	}
#endif
	if (ret == 0)
	    fp = fopen (path, mode);
	else
	    fp = NULL;
    }

    EFREE (path2);
    return fp;
}

void *
emalloc (int size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

void *
erealloc (void *ptr, size_t size)
{
    void *p = realloc (ptr, size);
    if (p == NULL) perror ("realloc");
    return p;
}

void
efree (void *ptr, char *file, int line)
{
    assert (ptr != NULL);
    free (ptr);
}

char *
estrdup (const char *s)
{
    assert (s != NULL);
#if 0
    char *p = strdup (s);
#else
    int size = strlen (s) + 1;
    char *p = emalloc (size);
    memcpy (p, s, size);
#endif
    if (p == NULL) perror ("strdup");
    return p;
}

#if 0
int
easprintf (char **sp, const char *fmt, ...)
{
    int ret;
    va_list pvar;
    va_start (pvar, fmt);
    ret = vasprintf (sp, fmt, pvar);
    va_end (pvar);
    if (ret == -1) perror ("asprintf");
    return ret;
}
#else
int
easprintf (char **sp, const char *fmt, ...)
{
    int ret, size;
    va_list pvar;
    va_start (pvar, fmt);
    size = vsnprintf (NULL, 0, fmt, pvar);
    va_end (pvar);
    if (size == -1) {
	perror ("vsnprintf"); exit (1);
    }

    va_start (pvar, fmt);
    *sp = emalloc (size + 1);
    ret = vsnprintf (*sp, size + 1, fmt, pvar);
    va_end (pvar);
    if (ret == -1) {
	perror ("vsnprintf"); exit (1);
    }

    return ret;
}
#endif

char *
comment_to_space (char *p)
{
    char *s = estrdup (p), *head = s;

    while (*s != '\0') {
	if (s [0] == '/' && s [1] == '*') {

	    s [0] = s [1] = ' ';
	    s += 2;

	    while (1) {
		if (s [0] == '\0')
		    return head;
		if (s [0] == '*' && s [1] == '/') {
		    s [0] = s [1] = ' ';
		    s += 2;
		    break;
		}
		if (s [0] == '\n') {
		    /* preserving newline */
		    s++;
		} else {
		    s [0] = ' ';
		    s++;
		}
	    }
	}
	if (s [0] == '/' && s [1] == '/') {
	    while (1) {
		if (s [0] == '\0')
		    return head;
		if (s [0] == '\n') {
		    /* preserving newline */
		    s++;
		    break;
		}
		s [0] = ' ';
		s++;
	    }
	}
	if (s [0] == '\0')
	    return head;
	s++;
    }
    return head;
}

char *
str_to_xml (char *s)
{
    char *p, *q;
    char *xml;
    int size;

    assert (s != NULL);
    size = 0;
    for (p = s; *p != '\0'; p++) {
	switch (*p) {
	case '<':  size += strlen ("&lt;");   break;
	case '>':  size += strlen ("&gt;");   break;
	case '&':  size += strlen ("&amp;");  break;
	case '\'': size += strlen ("&apos;"); break;
	case '\"': size += strlen ("&quot;"); break;

/* To cope with "3.3.3 attribute-value normalization" in XML spec., 
   where each CR, LF or tab is converted to a space. */
	case '\n': size += strlen ("&#x0A;"); break;
	case '\r': size += strlen ("&#x0D;"); break;
	case '\t': size += strlen ("&#x09;"); break;
	case '\f': size += strlen (ESC_FORM_FEED); break;
	default:   size++;                    break;
	}
    }
    xml = emalloc (size + 1);

    q = xml;
    for (p = s; *p != '\0'; p++) {
	switch (*p) {
	case '<':  strcpy (q, "&lt;");   q += strlen ("&lt;");   break;
	case '>':  strcpy (q, "&gt;");   q += strlen ("&gt;");   break;
	case '&':  strcpy (q, "&amp;");  q += strlen ("&amp;");  break;
	case '\'': strcpy (q, "&apos;"); q += strlen ("&apos;"); break;
	case '\"': strcpy (q, "&quot;"); q += strlen ("&quot;"); break;

	case '\n': strcpy (q, "&#x0A;"); q += strlen ("&#x0A;"); break;
	case '\r': strcpy (q, "&#x0D;"); q += strlen ("&#x0D;"); break;
	case '\t': strcpy (q, "&#x09;"); q += strlen ("&#x09;"); break; 
	case '\f': strcpy (q, ESC_FORM_FEED);
	                                 q += strlen (ESC_FORM_FEED); break; 
	default:   *q = *p;              q++;                    break;
	}
    }
    *q = '\0';
    return xml;
}

char *
str_to_xml2 (char *s)
{
    char *p = comment_to_space (s);
    char *q = str_to_xml (p);
    EFREE (p);
    return q;
}

void
xml_encode (FILE *fp, char *text)
{
    char *p;

    for (p = text; *p != '\0'; p++) {
	switch (*p) {
	case '<':  fputs ("&lt;", fp);   break;
	case '>':  fputs ("&gt;", fp);   break;
	case '&':  fputs ("&amp;", fp);  break;
	case '\'': fputs ("&apos;", fp); break;
	case '\"': fputs ("&quot;", fp); break;
	default:   fputc (*p, fp);       break;
	}
    }
}

#if 0
/* add the option '-Wl,-wrap=malloc' to gcc */
void *
__wrap_malloc (size_t size)
{ 
    return __real_malloc (size);
}

void *
__wrap_realloc (void *ptr, size_t size)
{
    return __real_realloc (ptr, size);
}

void
__wrap_free (void *ptr)
{
    __real_free (ptr);
}
#endif

/* ==================================================================== */
