#include <stdio.h>
/* tifdef's src2html - a tiny converter from source code to HTML-tagged code
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (C) 2003, 2006.  gondow(at)cs.titech.ac.jp, All rights reserved.
  $Id: tifdef.c,v 1.2 2006/06/17 16:37:48 gondow Exp $
 */
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>

#define DEFAULT_DIR	"./HTML"
static char *char_conv (int c);
static void make_dir (char *dirname);
static void open_xml (char *filename);
static void close_xml (void);
static void src2html (char *filename);

static int tifdef_lineno_from, tifdef_lineno_to;

/* ================================================================== */
static char *
char_conv (int c)
{
    switch (c) {
    case '<':  return "&lt;";
    case '>':  return "&gt;";
    case '&':  return "&amp;";
    default:   assert (0); return NULL;
    }
}

static void
make_dir (char *dirname)
{
    char buf [BUFSIZ];
    char *dir = strdup (dirname);
    char *p   = strrchr (dir, '/');
    *p = '\0';
    snprintf (buf, sizeof (buf), "mkdir -p %s", dir);
    system (buf);
    free (dir);
}

static void
set_tifdef (FILE *fp)
{
    int ret;
    char *ret2, buf [BUFSIZ];
    
    ret2 = fgets (buf, sizeof (buf), fp);
    if (ret2 == NULL) {
	tifdef_lineno_from = INT_MAX;
	tifdef_lineno_to   = INT_MAX;
	return;
    }
    ret = sscanf (buf, "%d %d", &tifdef_lineno_from, &tifdef_lineno_to);
    if (ret != 2) {
	tifdef_lineno_from = INT_MAX;
	tifdef_lineno_to   = INT_MAX;
	return;
    }
}

static void
src2html (char *filename)
{
    char *xml_filename, *dir, *tifdef_cmd;
    char buf [BUFSIZ];
    FILE *fp_in, *fp_out, *fp_xml, *fp_tifdef;
    int i, c, prev_c = '\n';

    snprintf (buf, sizeof (buf), "%s/%s.html", DEFAULT_DIR, filename);

    make_dir (buf);

    fp_in  = fopen (filename, "r");
    if (fp_in  == NULL) {
	fprintf (stderr, "file = %s\n", filename);
	perror ("fopen (r)");
	exit (1); 
    }
    fp_out = fopen (buf, "w");
    if (fp_out == NULL) {
	fprintf (stderr, "file = %s\n", buf);
	perror ("fopen (w)");
	exit (1);
    }

    asprintf (&xml_filename, "%s.xml", filename);
    fp_xml = fopen (xml_filename, "r");

    dir = getenv ("TBCPPA_DIR");
    if (dir == NULL)
	dir = "./TBCppA";

    asprintf (&tifdef_cmd, "xsltproc %s/../tifdef.xsl %s | sort -n",
	      dir, xml_filename);
    free (xml_filename);
    fp_tifdef = popen (tifdef_cmd, "r");
    if (fp_tifdef == NULL) {
	fprintf (stderr, "tifdef: error: %s\n", tifdef_cmd);
	perror ("popen");
	exit (1);
    }
    free (tifdef_cmd);
    
    set_tifdef (fp_tifdef);

    fprintf (fp_out, "<html><body>\n<h1>%s</h1>"
	     "<a name=\"top\"/>\n<hr/>\n<pre>\n", filename);

    i = 1;
    while ((c = getc (fp_in)) != EOF) {
	if (prev_c == '\n') {
	    if ((tifdef_lineno_from <= i - 1) && (i - 1 < tifdef_lineno_to)) {
		fprintf (fp_out, "</span>");
	    }
	    fprintf (fp_out, "<a name=\"%d\"><font color='blue'>%5d</font> ",
		     i, i);
	    if (i == tifdef_lineno_to)
		set_tifdef (fp_tifdef);
	    if (tifdef_lineno_from <= i && i < tifdef_lineno_to) {
		fprintf (fp_out,
			 "<span style='background:grey; color:black'>");
	    }
	    i++;
	}
	if ((c == '<') || (c == '>') || (c == '&')) {
	    fprintf (fp_out, "%s", char_conv (c));
	} else {
	    putc (c, fp_out);
	}
	prev_c = c;
    }
    fprintf (fp_out, "</pre><hr/></body></html>\n");

    fclose (fp_in);
    fclose (fp_out);
    fclose (fp_xml);
    fclose (fp_tifdef);
}
/* ================================================================== */
int
main (int argc, char *argv [])
{
    int i;

    if (argc <= 1) {
	fprintf (stderr, "Usage: %s filename ...\n", argv [0]);
	exit (1);
    }

    for (i = 1; i < argc; i++) {
	src2html (argv [i]);
    }
}
/* ================================================================== */
