/* txml: converter from @file to XML file
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: txml.l,v 1.5 2006/10/18 11:14:24 gondow Exp gondow $ */

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "misc.h"

#define DIR "./TBCppA"
#define DOUBLE_SUFFIX "_1528552"
#define XML_PREAMBLE                             \
    "<?xml version='1.0'?>\n"                    \
    "<!DOCTYPE TBCppA SYSTEM \"TBCppA.dtd\"\n"   \
    "[\n"                                        \
    "    <!ENTITY db_file SYSTEM '%s/%s'>\n"     \
    "    <!ENTITY comm_file SYSTEM '%s/%s'>\n"   \
    "]>\n"                                       \
    "<!--\n"                                     \
    "<!DOCTYPE TBCppA SYSTEM \"TBCppA.dtd\"\n"   \
    "[\n"                                        \
    "    <!ENTITY db_file SYSTEM '%s/%s'>\n"     \
    "    <!ENTITY comm_file SYSTEM '%s/%s'>\n"   \
    "]>\n"                                       \
    "-->\n"                                      \
"<TBCppA>\n"
#define XML_POSTAMBLE                            \
    "<filename_list>&db_file;</filename_list>\n" \
    "<lexical_info>&comm_file;</lexical_info>\n" \
    "</TBCppA>\n"

enum {
    PATH_MAX = 1024,
};

/* ==================================================================== */
int yylex (void);
/* ==================================================================== */
/* ==================================================================== */
static char db_file [PATH_MAX];
static char comm_file [PATH_MAX];
static int preamble_seen = 0;
static char *dir;
/* ==================================================================== */
%}

%option nounput
%x string character

O      [0-7]
D      [0-9]
H      [a-fA-F0-9]
L      [a-zA-Z_$]
WS     [ \t\f\v\r]           
L_IDENTIFIER ({L}|{L_UNIVERSAL_CHARACTER_NAME})({L}|{D})*

L_SIMPLE_ESCAPE_SEQUENCE \\[ntvbrfa\\\?'"]
L_OCTAL_ESCAPE_SEQUENCE  \\{O}{1,3}
L_HEXADECIMAL_SEQUENCE   \\(x|X){H}+
L_UNIVERSAL_CHARACTER_NAME \\(u{H}{4}|U{H}{8})
L_ESCAPE_SEQUENCE        {L_SIMPLE_ESCAPE_SEQUENCE}|{L_OCTAL_ESCAPE_SEQUENCE}|{L_HEXADECIMAL_SEQUENCE}|{L_UNIVERSAL_CHARACTER_NAME}

%% /* ==================================================================== */
^@@.*$        { /* @@preamble db_file.xml foo.comm */
                  int ret;
		  if (preamble_seen) return 1;
		  ret = sscanf (yytext, "@@preamble %s %s",
				db_file, comm_file);
		  if (ret != 2) {
		      fprintf (stderr, "txml: bad preamble format: %s\n",
			       yytext);
		      exit (1);
		  }
		  fprintf (yyout, XML_PREAMBLE,
			   dir, db_file, dir, comm_file,
			   dir, db_file, dir, comm_file);
		  preamble_seen = 1;
              }
@\"\/[^\"]+\" { /* @"/foo"  -> </foo> */
                  chop (yytext); 
                  fprintf (yyout, "</%s>", &yytext [3]);
              }
@\"[^\"]+\/\" { /* @"foo/" -> <foo/> */
                  chop (yytext); chop (yytext); 
		  fprintf (yyout, "<%s/>", &yytext [2]);
              }
@\"[^\"]+\"   { /* @"foo" -> <foo> */
                  chop (yytext); 
		  fprintf (yyout, "<%s>", &yytext [2]);
              }

{L_IDENTIFIER}_1528552 {
                  yytext [strlen(yytext) - strlen(DOUBLE_SUFFIX)] = '\0';
		  fprintf (yyout, "%s", yytext);
              }

L\"    	       { BEGIN(string); fprintf (yyout, "L&quot;"); }
\"    	       { BEGIN(string); fprintf (yyout, "&quot;"); }
<string>"<"    { fprintf (yyout, "&lt;"); }
<string>">"    { fprintf (yyout, "&gt;"); }
<string>"&"    { fprintf (yyout, "&amp;"); }
<string>"'"    { fprintf (yyout, "&apos;"); }
<string>"\\\\" { fprintf (yyout, "%s", yytext); }
<string>"\\\"" { fprintf (yyout, "\\&quot;"); }
<string>"\""   { BEGIN(INITIAL); fprintf (yyout, "&quot;"); }
<string>.      { fprintf (yyout, "%s", yytext); }

L'                  { BEGIN(character); fprintf (yyout, "L&apos;"); }
'                   { BEGIN(character); fprintf (yyout, "&apos;"); }
<character>"\""     { fprintf (yyout, "&quot;"); }
<character>"<"      { fprintf (yyout, "&lt;"); }
<character>">"      { fprintf (yyout, "&gt;"); }
<character>"&"      { fprintf (yyout, "&amp;"); }
<character>[^\\\n'] { fprintf (yyout, "%s", yytext); }
<character>"\\\""   { fprintf (yyout, "\\&quot;"); }
<character>"\\'"    { fprintf (yyout, "\\&apos;"); }
<character>{L_ESCAPE_SEQUENCE} { fprintf (yyout, "%s", yytext); }
<character>'        { BEGIN(INITIAL); fprintf (yyout, "&apos;"); }


"<"           { fprintf (yyout, "&lt;"); }
">"           { fprintf (yyout, "&gt;"); }
"&"           { fprintf (yyout, "&amp;"); }

^{WS}*#.*\n  { fprintf (stderr, "CPP directive removed: %s\n", yytext); }

<<EOF>>       { fprintf (yyout, XML_POSTAMBLE); return 0; }

.             { fputc (yytext [0], yyout); }
%% /* ==================================================================== */

int
main (int argc, char *argv [])
{

    if (argc != 1) {
	fprintf (stderr, "Usage: %s < foo.E > foo.xml\n", argv [0]);
	fprintf (stderr, "You need recompile to change DOUBLE_SUFFIX: %s\n",
		 DOUBLE_SUFFIX);
	exit (1);
    } 

    dir = getenv ("TBCPPA_DIR");
    if (dir == NULL)
	dir = DIR;

    while (yylex () != 0)
	;
    return 0;
}
/* ==================================================================== */
